/**
 * $Id: Actor.java,v 1.20 2002/04/03 00:43:34 cvarela Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;

import java.lang.reflect.Method;

public class Actor extends Thread implements java.io.Serializable {
	private Mailbox mailbox;

	public Actor() {
		mailbox = new Mailbox(this);
		start();
	}

	public void send(Message message) {
		RunTime.receivedMessage();
		mailbox.put(message);
	}

	// package default access
	synchronized void process(Message message) {
		// cvarela:  finishedProcessingMessage *should* be called when
		// "process" terminates abnormally (by throwing an exception).
		Method method = message.getMethod();
		Object[] args = message.getArguments();
		Object token = null;
		try {
			token = method.invoke(this, args);
		} catch (java.lang.reflect.InvocationTargetException ite){
			Throwable targetException = ite.getTargetException();
			if (targetException instanceof ThreadDeath)
				throw new ThreadDeath();
			else {
				System.err.println(targetException+" processing message:"+method);
				targetException.printStackTrace();
				// rethrow it to source actor's complaint box
				// throw targetException;
			}
		} catch (IllegalAccessException iae) {
			System.err.println(iae+" processing message:"+method);
		}
		Message continuation = message.getContinuation();
		if (continuation != null) {
			if (continuation.getWithToken())
				try {
					continuation.setToken(token);
				} catch (NoTokenInContinuationException ntice){
					System.err.println("SALSA Internal Error:"+ntice);
				}
			continuation.getTarget().send(continuation);
		}
		RunTime.finishedProcessingMessage();
	}

	public void run() {
		while (true) {
			Message nextMessage = mailbox.get();

			// Tries to get Actor's lock, only after getting the message.
			process(nextMessage);
		}
	}

	/**
	 * *** This is dangerous to call ***
	 *
	 * As soon as the mailbox becomes empty, the actor dies.  There may
	 * be other messages en route coming from other actors; which will
	 * become lost.
	 *
	 * This method should only be called, when we are absolutely sure that
	 * no more messages are coming to this actor from anywhere.  It is
	 * provided to be able to exit the Java Virtual Machine.
	 *
	 * As of Salsa 0.2. this method has been deprecated.
	 */
	public void die() {
		if (!mailbox.isEmpty()) {
			synchronized(this) {
				try {
					wait();  // The lock of Actor means its mailbox is not empty.
				} catch (InterruptedException ie) {
					System.err.println("Dying:"+ie);
				}
			}
		}
		synchronized(this) {  // To prevent dying while processing a message
			stop();
		}
	}

	protected synchronized void notifyEmptyMailbox() {
		notify();
	}

	public String toString() {
		return getClass().getName()+":"+super.toString();
	}

	/**
	 * The Salsa compiler should always generate code that uses the  
	 * version of this method:  
	 * 		send(Message message)
	 *
	 * The following methods demonstrate the type of code that gets
	 * generated by the SALSA compiler.
	 */

	/* Self-sent methods with no parameters or continuations. */
	// Code equivalent to SALSA message send:
	//
	//   self:methodName();
	//
	// or just:
	//
	//   methodName();
	//
	public void send(String methodName) {
		try {
			send(Message.createMessage(this, this, methodName));
		} catch (NoSuchMessageException nsme) {
			System.err.println("No such message with no arguments: "+methodName);
		}
	}

	/**
	 * Self-sent methods with no parameters and self-sent continuations 
	 *  with no parameters.
	 */
	// Code equivalent to SALSA message and continuation:
	//
	//   self:methodName() -> self:continuationName();
	//
	// Or just:
	//
	//   methodName() -> continuationName();
	//
	public void send(String methodName, String continuationName) {
		try {
			send(Message.createMessage(this, this, methodName, this, continuationName));
		} catch (NoSuchMessageException nsme) {
			System.err.println("No such message with no arguments: "+nsme.getMessage());
		}
	}


	/**
	 * Methods with no parameters and two continuations with no parameters. 
	 * The actors to be targeted are also arguments to this method.
	 */
	// Code equivalent to SALSA message and continuation:
	//
	// a0:methodName0() -> a1:methodName1() -> a2:methodName2();
	//
	public void send(Actor a0, String methodName0, Actor a1, String methodName1, Actor a2, String methodName2) {
		try {
			Actor[]  targets = new Actor[3];
			String[] methodNames = new String[3];
			targets[0] = a0;  targets[1] = a1; targets[2] = a2;
			methodNames[0] = methodName0;
			methodNames[1] = methodName1;
			methodNames[2] = methodName2;
			a0.send(Message.createMessage(this, targets, methodNames));
		} catch (NoSuchMessageException nsme) {
			System.err.println("No such message with no arguments:"+nsme.getMessage());
		} catch (MessageCreationException mce) {
			System.err.println("SALSA Internal Error:"+mce.getMessage());
		}   
	}

	/**
	 * Methods with no parameters and continuations with 
	 * one parameter: token. 
	 */
	// Code equivalent to SALSA message and continuation:
	//
	// a0:methodName0() -> a1:methodName1(token)
	//
	public void send(Actor a0, String methodName0, Actor a1, String methodName1, boolean passingToken) {
		try {
			Actor[]  targets = new Actor[2];
			String[] methodNames = new String[2];
			Object[][] arguments = new Object[2][];
			int[] tokens = new int[2];
			targets[0] = a0;  targets[1] = a1;
			methodNames[0] = methodName0;
			methodNames[1] = methodName1;
			arguments[0] = new Object[0];
			arguments[1] = new Object[1];
			tokens[0] = -1;
			tokens[1] = 0;
			a0.send(Message.createMessage(this, targets, methodNames, arguments, tokens));
		} catch (NoSuchMessageException nsme) {
			System.err.println("No such message:"+nsme.getMessage());
			nsme.printStackTrace();
		} catch (MessageCreationException mce) {
			System.err.println("SALSA Internal Error:"+mce.getMessage());
		}   
	}

	/**
	 * Method with one "int" parameter and no continuations.
	 */
	// Code equivalent to SALSA message:
	//
	// int parameter;
	// a:methodName(parameter);
	//
	public void send(Actor a, String methodName, int parameter) {
		try {
			Actor[]  targets = new Actor[1];
			String[] methodNames = new String[1];
			Object[][] arguments = new Object[1][];
			arguments[0] = new Object[1];
			targets[0] = a; 
			methodNames[0] = methodName;
			arguments[0][0] = new Integer(parameter);
			a.send(Message.createMessage(this, targets, methodNames, arguments));
		} catch (NoSuchMessageException nsme) {
			System.err.println(nsme);
		} catch (MessageCreationException mce) {
			System.err.println("SALSA Internal Error:"+mce.getMessage());
		}   
	}

//     private void writeObject(java.io.ObjectOutputStream out)
// 	throws IOException{
// 	// the no-args constructor of the non-serializable Thread class
// 	// gets called at "readObject" time.
//     }

//     private void readObject(java.io.ObjectInputStream in)
// 	throws java.io.IOException, ClassNotFoundException{
// 	// here we only need to make sure the thread is "restarted".
// 	in.defaultReadObject();
// 	start();  // done at RMSPDaemon.onReceive()
//     }
}
