/**
 * This is the Message class.
 */

package salsa.language;

import java.lang.reflect.Method;

public class Message implements java.io.Serializable {
	private Actor source;
	private Actor target;
	transient private Method method;
	private Object[] arguments;
	private Message continuation;
	private int tokenPosition;		// token position in continuation arguments
	private boolean withMessage;	// pass itself as first argument to
									// method invocation.

	/**
	 * This constructor is not public.  It can only be called from 
	 * classes in the salsa.language package.
	 */
	Message(Actor source, Actor target, Method method, Object[] arguments, Message continuation, int tokenPosition, boolean withMessage) {
		this.source = source;
		this.target = target;
		this.method = method;
		this.arguments = arguments;
		this.continuation = continuation;
		this.tokenPosition = tokenPosition;
		this.withMessage = withMessage;
	}

	/**
	 * No token being passed to continuation.  -1 represents such case.
	 */
	Message(Actor source, Actor target, Method method, Object[] arguments, Message continuation) {
		this(source, target, method, arguments, continuation, -1, false);
	}

    private void setSource(Actor source) {
	this.source = source;
    }

    private void setTarget(Actor target) {
	this.target = target;
    }

    public Actor getSource() {
	return source;
    }

    public Actor getTarget() {
	return target;
    }

    public Method getMethod() {
	return method;
    }

    public Object[] getArguments() {
	if (getWithMessage())
	    setMessage();
	return arguments;
    }

    public Message getContinuation() {
	return continuation;
    }

    public void setContinuation(Message continuation) {
	this.continuation = continuation;
    }

    public boolean getWithToken() {
	return (tokenPosition != -1);
    }

    public void setToken(Object token) throws NoTokenInContinuationException {
	try {
	    arguments[tokenPosition] = token;
	} catch (ArrayIndexOutOfBoundsException e) {
	    throw new NoTokenInContinuationException();
	}
    }

    public boolean getWithMessage() {
	return (withMessage);
    }

    public void setMessage() {
	arguments[0] = this;
    }

    public void setLastContinuation(Message continuation) {
	if (this.continuation == null)
	    this.continuation = continuation;
	else
	    continuation.setLastContinuation(continuation);
    }

    /*Helper message constructors for common compiler generated cases */
    static Message createMessage(Actor source, Actor target, String methodName, Object[] arguments, Message continuation, int tokenPosition,  Class tokenClass, boolean noPrimitive) throws NoSuchMessageException {
	Class[] parameterTypes = new Class[arguments.length];
	boolean withMessage = false;
	if (tokenPosition < -1){  
	    // Represents passing Message as argument.
	    // -2,-4,-5,..,-n-2 --> tokenPosition = -1,1,2,..,n-1
	    withMessage = true;
	    parameterTypes[0] = Message.class;
	    tokenPosition = -tokenPosition - 3;
	}
	if (tokenPosition == -1){
	    for (int i = (withMessage? 1 : 0); i < arguments.length; i++){
                if (noPrimitive)
                    parameterTypes[i] =
                        arguments[i].getClass();
                else
		    // Disallows the use of Wrapper objects as parameters.
		    // Is this a good idea?  
		    // e.g., print(Integer) not allowed, instead print(int)
		    parameterTypes[i] = 
		        convertToPrimitiveType(arguments[i].getClass());
	    }
	} else {
	    // cvarela:  
	    // We use returnType of message originating the continuation.  
	    // We'll consider primitive types later.
	    for (int i = (withMessage? 1 : 0); i < arguments.length; i++){
		if (tokenPosition == i)
		    if (noPrimitive)
			parameterTypes[i] = tokenClass;
		    else
			parameterTypes[i] = 
			    convertToPrimitiveType(tokenClass);
		else
                    if (noPrimitive)
                        parameterTypes[i] =
                            arguments[i].getClass();
                    else
                        parameterTypes[i] =
                            convertToPrimitiveType(arguments[i].getClass());
	    }
	}
	try {
	    Method method =
		mostSpecific(target.getClass(), methodName, parameterTypes);
	    return new 
		Message(source,
			target,
			method,
			arguments, 
			continuation,
			tokenPosition,
			withMessage);
	} catch (NoSuchMethodException nsme){
            if (!noPrimitive)
                    return createMessage(source, target, methodName, arguments,
                                         continuation, tokenPosition, tokenClass,
                                         true);
            else 
 	        throw new NoSuchMessageException
		    (methodSignature(methodName, parameterTypes) +
		     " in "+ target.getClass());
	}	
    }

      static Message createMessage(Actor source, Actor target,
                                 String methodName, Object[] arguments,
                                 Message continuation,
                                 int tokenPosition,  Class tokenClass)
        throws NoSuchMessageException {
        return createMessage(source, target, methodName, arguments, continuation,
                             tokenPosition, tokenClass, false);
      }

    // SALSA code:
    //
    // target:methodName();
    //
    static Message createMessage(Actor source, Actor target, String methodName) throws NoSuchMessageException {
	return createMessage(source,
			     target,
			     methodName,
			     new Object[0],
			     null,
			     -1,
			     null);
    }

    // SALSA code:
    //
    // target:methodName() -> continuationTarget:continuationMethodName();
    //
    static Message createMessage(Actor source, 
				 Actor target, 
				 String methodName,
				 Actor continuationTarget, 
				 String continuationMethodName) 
	throws NoSuchMessageException{
	Object[] arguments = new Object[0];
	Message continuation =
	    createMessage(target,
			  continuationTarget,
			  continuationMethodName,
			  arguments,
			  null,
			  -1,
			  null);
	return createMessage(source,
			     target,
			     methodName,
			     arguments,
			     continuation,
			     -1,
			     null);
    }

    // SALSA source actor code:
    //
    // a0:methodName0() -> a1:methodName1() -> a2:methodName2() ...
    //
    // where targets[i] = ai, methodNames[i] = methodNamei
    //   
    static Message createMessage(Actor source,
				 Actor[] targets, 
				 String[] methodNames)
	throws NoSuchMessageException, MessageCreationException{
	Object[][] arguments = new Object[targets.length][];
	int[] tokens = new int[targets.length];
	for (int i = 0; i < targets.length; i++){
	    arguments[i] = new Object[0];
	    tokens[i] = -1;
	}
	return createMessage(source, targets, methodNames, arguments, tokens, null, 0);
    }

    // SALSA source actor code:
    //
    // a0:methodName0(arg00, arg01,..arg0n0) 
    // -> a1:methodName1(arg10, arg11,..arg1n1) 
    // -> a2:methodName2(arg20, arg21,...) ...
    //
    // where targets[i] = ai, methodNames[i] = methodNamei, and
    //       arguments[i] = [argi0, argi1,... argini]
    //   
    static Message createMessage(Actor source,
				 Actor[] targets, 
				 String[] methodNames,
				 Object[][] arguments)
	throws NoSuchMessageException, MessageCreationException{
	int[] tokens = new int[targets.length];
	for (int i = 0; i < targets.length; i++){
	    tokens[i] = -1;
	}
	return createMessage(source, targets, methodNames, arguments, tokens, null, 0);
    }

    // SALSA source actor code:
    //
    // a0:methodName0(arg00, arg01,..arg0n0)  token0 == arg0t0
    // -> a1:methodName1(arg10, arg11,..arg1n1)  token1 == arg1t1
    // -> a2:methodName2(arg20, arg21,...) ...
    //
    // where targets[i] = ai, methodNames[i] = methodNamei, and
    //       arguments[i] = [argi0, argi1,... argini], and
    //       tokens[i] = ti
    //   
    public static Message createMessage(Actor source,
					Actor[] targets, 
					String[] methodNames,
					Object[][] arguments,
					int[] tokens)
	throws NoSuchMessageException, MessageCreationException{
	return createMessage(source, 
			     targets, 
			     methodNames, 
			     arguments, 
			     tokens, 
			     null,  // no token passed initially
			     0);
    }

    public static Message createMessage(Actor source,
					Actor[] targets, 
					String[] methodNames,
					Object[][] arguments,
					int[] tokens,
					Class tokenClass,
					int currentIndex)
	throws NoSuchMessageException, MessageCreationException{
	try {
	    if (currentIndex == targets.length)
		return null;
	    else {
		Message message = 
		    createMessage(source, 
				  targets[currentIndex],
				  methodNames[currentIndex],
				  arguments[currentIndex],
				  null, // continuation set later
				  tokens[currentIndex],
				  tokenClass);
		Message continuation =
		    createMessage(targets[currentIndex], 
				  targets, 
				  methodNames,
				  arguments,
				  tokens,
				  message.getMethod().getReturnType(),
				  currentIndex+1);
		message.setContinuation(continuation);
		return message;
	    }
	} catch (ArrayIndexOutOfBoundsException e){
	    throw new MessageCreationException
		("Different number of targets, method names, arguments or tokens");
	}
    }

    /**
     * Converts Wrapper classes to their corresponding primitive  types:
     * boolean, byte, char, short, int, long, float, and double, and void. 
     */   
    protected static Class convertToPrimitiveType(Class c){
	// James' faster version of salsa's convertToPrimitiveType method
	// jiw 6 Aug 99
	try {
	    return  (Class) c.getField("TYPE").get(null);
	} catch (Exception n) {
	}
	return c;
    }

    /** 
     * Find most specific method to apply according to parameter types.
     */
    /* cvarela: left to right.  Should avoid ambiguous method calls...
     * and should be decided at COMPILATION time, not at RUN-TIME.
     * See:  http://java.sun.com/docs/books/jls/html/15.doc.html
     */
    private static Method mostSpecific(Class clazz, 
				       String methodName,
				       Class[] parameterTypes)
	throws NoSuchMethodException {
	try {
	    return clazz.getMethod(methodName, parameterTypes);
	} catch (NoSuchMethodException nsme){
//	    if (parameterTypes.length > 1)
//		System.err.println("Salsa Warning: May not be choosing"+
//				   " most specific method for:"+
//				   methodSignature(methodName, parameterTypes));
	    return mostSpecific(clazz, methodName, parameterTypes, 0);
	}
    }

    private static Method mostSpecific(Class clazz, 
				       String methodName,
				       Class[] parameterTypes,
				       int currentIndex)
	throws NoSuchMethodException {
	try {
	    return clazz.getMethod(methodName, parameterTypes);
	} catch (NoSuchMethodException nsme){
	    try {
		Class type = parameterTypes[currentIndex];
		while (parameterTypes[currentIndex] != null){
		    parameterTypes[currentIndex] = 
			parameterTypes[currentIndex].getSuperclass();
		    try{
			return clazz.getMethod(methodName, parameterTypes);
		    } catch (NoSuchMethodException nsme2){
			continue;
		    }
		}
		parameterTypes[currentIndex] = type;
		return mostSpecific(clazz, 
				    methodName, 
				    parameterTypes, 
				    currentIndex + 1);
	    } catch (ArrayIndexOutOfBoundsException aiobe){
//		if (parameterTypes.length > 1)
//		    System.err.println("Salsa Warning: "+
//				       "Method lookup not finished. "+
//				       "There may be a method for: "+
//				       methodSignature(methodName, parameterTypes));
		throw new NoSuchMethodException();
	    }
	}
    }

    private static String methodSignature(String methodName, 
					  Class[] parameterTypes){
	StringBuffer typeNames = new StringBuffer();
	for (int i = 0; i < parameterTypes.length; i++)
	    try {
		typeNames.append(parameterTypes[i].getName());
	    } catch (NullPointerException npe){
		typeNames.append("null");
		continue;
	    }
	return (methodName+"("+typeNames+")");
    }

    // only for wwc.messaging package access
    public void updateReferences(){
	UniversalActor sourceActor = 
	    ((UniversalActor)getSource()).getRemoteReference();
	UniversalActor targetActor =
	    ((UniversalActor)getTarget()).getRemoteReference();
	if (sourceActor != null) 
	    setSource(sourceActor);
	if (targetActor != null)
	    setTarget(targetActor);
	if (continuation != null)
	    continuation.updateReferences();
    }

    public String toString(){
	return ("[Source:"+source+
		", Target:"+target+
		", Method:"+method.toString()+
		", Arguments:"+arguments+
		", Continuation:"+continuation+
		", TokenPosition:"+tokenPosition+
		", WithMessage:"+withMessage);
    }

	private void writeObject(java.io.ObjectOutputStream out) throws java.io.IOException {
		out.defaultWriteObject();
		out.writeObject(method.getName());
		out.writeObject(method.getParameterTypes());
		out.flush();
	}

	private void readObject(java.io.ObjectInputStream in) throws java.io.IOException, ClassNotFoundException {
		in.defaultReadObject();
		String methodName = (String) in.readObject();
		Class[] parameterTypes = (Class[]) in.readObject();
		try {
			method = target.getClass().getMethod(methodName, parameterTypes);
		} catch (NoSuchMethodException e){
			throw new java.io.IOException("Unable to recreate Method "+methodName+": "+e);
		}
	}
}



