/**
 * A message group contains a set of messages, which are to be sent
 * asynchronously to a group of actors, an array of tokens returned
 * by those actors after processing the messages, and a continuation 
 * to be used after all messages have been processed.
 */

package salsa.language;

import java.lang.reflect.Array;

public class MessageGroup extends Actor {
	private Message[] messages;
	private Object[] tokens;
	private Message continuation;
	private int tokensSet;

	MessageGroup(Message[] messages, Message continuation) {
		this.messages = messages;
		this.tokens = new Object[messages.length];
		this.continuation = continuation;
		tokensSet = 0;
	}

	public void process() {
		for (int i = 0; i < messages.length; i++) {
			// messages[i].getTarget().send(messages[i])
			//  -> ack(i, token);
			try {
				Message continuation = null;
				{
					//  -> ack(i, token)
					Actor[] continuationTargets = {this};
					String[] continuationMethodNames = {"ack"};
					Object[][] continuationArguments = {{new Integer(i),
							null}};
					int[] continuationTokens = {1};
					continuation =
					Message.createMessage(this,
							continuationTargets,
							continuationMethodNames,
							continuationArguments,
							continuationTokens,
							Object.class,
							0);
				} 
				messages[i].setLastContinuation(continuation);
				messages[i].getTarget().send(messages[i]);
			} catch (NoSuchMessageException nsme) {
				nsme.printStackTrace();
			} catch (MessageCreationException mce) {
				System.err.println("SALSA Internal Error:"
						+mce.getMessage());
			}
		}
	}

	public void ack(Integer i, Object token) {
		tokens[i.intValue()] = token;
		if ((++tokensSet) == tokens.length) { // all tokens set
			if (continuation != null) {
				if (continuation.getWithToken())
					try {
						// convert array of Wrapper Objects 
						// to array of primitive types, if necessary.
						Class originalClass = tokens[0].getClass();
						Class primitiveType = Message.convertToPrimitiveType
						(originalClass);
						if (originalClass != primitiveType) {
							Object primitiveValuesArray = null;
							try {
								primitiveValuesArray = Array.newInstance(primitiveType,
										tokens.length);
								for (int j = 0; j < tokens.length; j++)
									Array.set(primitiveValuesArray,
											j, tokens[j]);
							} catch (IllegalArgumentException iae) {
								System.err.println(iae);
								iae.printStackTrace();
							} catch (RuntimeException rte) {
								System.err.println(rte);
								rte.printStackTrace();
							}
							continuation.setToken(primitiveValuesArray);
						} else {
							continuation.setToken(tokens);
						}
						// cvarela:  setArguments(tokens);  (better?)
						// it would be nice to do smart type checking here.
					} catch (NoTokenInContinuationException ntice) {
						System.err.println("SALSA Internal Error:"+ntice);
					}
				continuation.getTarget().send(continuation);
			}
			// this actor can now die, 
			// message counter doesn't decrease though..
			// die();
		} 
	}

	public void setContinuation(Message continuation) {
		this.continuation = continuation;
	}

	/**
	 * If the message group is homogeneous, getReturnType returns 
	 * an array of such homogeneous return type.  Otherwise,
	 * getReturnType returns an array of Object type.
	 */
	public Class getReturnType() {
		if (messages.length == 0)
			return null;
		Class defaultReturnType = messages[0].getMethod().getReturnType();
		for (int i=1; i<messages.length; i++) {
			if (defaultReturnType != messages[i].getMethod().getReturnType()) {
				defaultReturnType = Object.class;
				break; // return array of Objects type.
			}
		}
		// All methods return VOID.
		if (defaultReturnType == Void.TYPE)
			defaultReturnType = Object.class;
		return (Array.newInstance(defaultReturnType,0).getClass());
	}

	// SALSA source actor code:
	//
	// join ( targets[0]:methodNames[0](arguments[0]), ...
	//        targets[n-1]:methodNames[n-1](arguments[n-1]) )
	//      -> continuation;
	public static MessageGroup createMessageGroup(Actor source,
			Actor[] targets,
			String[] methodNames,
			Object[][] arguments,
			int[] tokenPositions,
			Message continuation)
			throws NoSuchMessageException {
		Message[] messages = new Message[targets.length];
		for (int i = 0; i < targets.length; i++) {
			messages[i] = Message.createMessage(source,
					targets[i],
					methodNames[i],
					arguments[i],
					null,
					tokenPositions[i],
					null);
		}
		return new MessageGroup(messages, continuation);
	}
}
