/**
 * $Id: UniversalActor.java,v 1.4 2002/01/16 00:35:30 tollr Exp $
 *
 * This is the Universal Actor class.  Universal actors can be
 * accessed remotely.  Clients need only have their UAN.
 * Universal actors can also migrate to an arbitrary remote location
 * described by a UAL.
 *
 * The host in the UAN needs to be running a UAN Daemon.  The host
 * specified in the UAL needs to be running an RMSP Daemon.
 *
 * An RMSP Daemon is responsible for keeping track of actors hosted
 * in its internal Stage.  It may also be used as a proxy to communicate
 * with actors hosted by AppletStages served by an HTTP server in the 
 * same host as the RMSP daemon.
 * 
 * Serialization of an actor passes an Actor Reference, i.e. an equivalent
 * actor with its "local" flag set to false.  The actor being serialized
 * remains in the current stage.  So, actors are passed BY REFERENCE.
 * 
 * Migration of an actor moves the Actor to a new Stage, leaving behind
 * an Actor Reference (forwarder).  The process also updates the UAN
 * daemon corresponding to the actor's name with the new location.
 * All internal actor references are serialized (passed by reference).
 *
 * Anonymous actors are allowed (no UAN), as well as named actors with
 * no current location (UAL).  Actors with no UAN and no UAL should
 * not be Universal.
 *
 * Author: Carlos Varela   cvarela@uiuc.edu   1999/11/13.
 */

package salsa.language;

import java.io.*;
import java.net.*;
import wwc.messaging.*;
import wwc.naming.*;

public class UniversalActor extends Actor {
	private UAN uan;
	private UAL ual;
	private boolean local;
	transient private UniversalActor remoteReference;

	public UniversalActor() {
		super();
		uan = null;
		ual = null;
		local = true;
		remoteReference = null;
	}

//     /* Used internally to return remote Actor references */
//     private UniversalActor(UAN uan, UAL ual){
// 	super();
// 	this.uan = uan;
// 	this.ual = ual;
// 	local = false;
// 	remoteReference = null;
//     }

	public UAN getUAN() {
		return uan;
	}

	public UAL getUAL() {
		return ual;
	}

	UniversalActor getRemoteReference() {
		return remoteReference;
	}

	/**
	 * Binds this object to a UAN and UAL so it can be located later.
	 *
	 * @param uan The universal actor name to bind this actor to.
	 * @param ual The universal actor locator to bind this actor to.
	 */
	public synchronized void bind(String uan, String ual) throws UANException, UALException {
		bind(new UAN(uan), new UAL(ual));
	}

	/**
	 * Binds this object to a UAN and UAL so it can be located later.
	 *
	 * @param uan The universal actor name to bind this actor to.
	 * @param ual The universal actor locator to bind this actor to.
	 */
	public synchronized void bind(UAN uan, UAL ual) throws UANException, UALException {
		this.uan = uan;
		this.ual = ual;
		try {
			// (1) moves "this" to UAL.
			moveTo(ual);
			// (2) update UANDaemon with new UAL.
			if (uan != null)  // not anonymous
				uan.updateLocation(ual);
		} catch (UnknownHostException e) {
			this.uan = null;
			this.ual = null;
			throw new UALException("RMSP Server "+ual.getHost()+" unreachable: "+e);
		} catch (IOException e) {
			this.uan = null;
			this.ual = null;
			throw new UALException("java.io.IOException: "+e);
		} catch (UANIOException e){
			this.uan = null;
			throw new UANException("Actor Name Server "+uan.getHost()+" unreachable: "+e);
		}
		local = false;
	}

//	public static UniversalActor byName(String uan) throws UANException {
//		return byName(new UAN(uan));
//	}

//	public static UniversalActor byLocation(String ual) throws MalformedUALException {
//		return byLocation(new UAL(ual));
//	}

//	public static UniversalActor byName(UAN uan) throws UANException {
//		UAL ual = uan.getLocation();
//		// need to get the real type of such actor to create proper
//		// instance of Actor reference...
//		return new UniversalActor(uan, ual);
//	}

//	public static UniversalActor byLocation(UAL ual) {
//		// Returns such anonymous Actor Reference
//		return new UniversalActor(null, ual);
//	}

	/** 
	 * These methods require the calling Actor to have the same type as 
	 * the actor corresponding to the UAN given.  
	 * For examples, see migration.Agent, migration.Migrate and
	 * migration.RemoteSend.
	 */
	public void getReferenceByName(String uan) throws UANException {
		getReferenceByName(new UAN(uan));
	}

	public void getReferenceByName(UAN uan) throws UANException {
		this.uan = uan;
		ual = uan.getLocation();
		local = false;
		remoteReference = null;
	}

	public void getReferenceByLocation(String ual) throws MalformedUALException {
		getReferenceByLocation(new UAL(ual));
	}

	public void getReferenceByLocation(UAL ual) {
		this.uan = null;
		this.ual = ual;
		local = false;
		remoteReference = null;
	}

	public void send(Message message) {
		if (((uan == null) && (ual == null)) || local) {
			// local actor:
			//System.out.println("Local send: "+message);
			super.send(message);
		} else {
			// remote actor:
			if (!RMSProtocol.usingMessengers) {
				try {
					//System.out.println("Remote send:"+message);
					Socket s = new Socket(ual.getHost(), ual.getPort());
					ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(s.getOutputStream()));
					out.writeObject(message);
					out.flush();
					out.close();
					s.close();
				} catch (UnknownHostException e) {
					System.err.println("Error sending message "+message+":"+e);
				} catch (IOException e) {
					System.err.println("Error sending message "+message+":"+e);
				}
			} else {  // with a messenger
				Messenger messenger = new Messenger(message);
				// SALSA:  messenger <- go();
				{
					Actor[] _targets = { messenger };
					String[] _methodNames = {"go"};
					Object[][] _arguments = {{}};
					int [] _tokens = {-1};
					try {
						messenger.send(Message.createMessage(null, _targets, _methodNames, _arguments, _tokens));
					} catch (NoSuchMessageException _nsme) {
						_nsme.printStackTrace();
					} catch (MessageCreationException _mce) {
						System.err.println("SALSA Internal Error:"+_mce.getMessage());
					}
				}
			}
		}
	}  

	synchronized void process(Message message) {
		if (!local) // Actor has moved with its mailbox!
			RunTime.finishedProcessingMessage();
		else
			super.process(message);
	}

	public synchronized void migrate(String newUAL) throws UALException, MigrationException {
		migrate(new UAL(newUAL));
	}

	public synchronized void migrate(UAL newUAL) throws MigrationException {
		if (!local)
			throw new MigrationException("Migration to "+newUAL+" failed:  Actor is not in this stage!");
		try {
			bind(uan, newUAL);
		} catch (Exception e) {
			throw new MigrationException(e.getMessage());
		}
	}

	void moveTo(UAL ual) throws UnknownHostException, IOException {
		Socket s = new Socket(ual.getHost(), ual.getPort());
		ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(s.getOutputStream()));
		out.writeObject(this);
		out.flush();
		out.close();
		s.close();
	}

// System environment actors (standardOutput, standardInput, etc)
// should not be serialized at run-time... they're static
// variables:  therefore they don't get serialized.  Need to be
// dynamically bound in the new stage though... not really, since
// new static actors get created at remote stage.
// System environment actors (standardOutput, standardInput, etc)
// should be dynamically bound at "read" time...
	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		// if UAN=UAL=null, dynamically generate UAL and 
		// bind actor to such UAL. i.e. it becomes remotely accessible
		if ((uan == null) && (ual == null)) { 
			// for now, just print a Warning:
			System.out.println("Warning: Trying to write a nonUniversal actor: "+this);
			// 	    try {
			// 		String localHost = InetAddress.getLocalHost().getHostName();
			// 		ual = new UAL("rmsp://"+localHost+":"+
			// 			      wwc.messaging.RMSPDaemon.getListeningPortNumber()+
			// 			      "/"+this.getName());
			// 		// cvarela: UAL may be already used...  
			// 		//          or RMSP daemon may be inactive...
			// 		moveTo(ual);  // anonymous messenger
			// 	    } catch (Exception e) {
			// 		e.printStackTrace();
			// 	    }
		}
		out.writeObject(uan);
		out.writeObject(ual);
		out.writeBoolean(false);
		out.flush();
	}

	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
		uan = (UAN) in.readObject();
		ual = (UAL) in.readObject();
		local = in.readBoolean();
		// if local in registry, replace "this" to *local* Actor.
		String localHost = InetAddress.getLocalHost().getHostName();
		if ((ual != null) && (ual.getHost().equals(localHost)) && (ual.getPort() == wwc.messaging.RMSPDaemon.getListeningPortNumber())) {
		// Reading a *local* actor.
		local = true;
			if (wwc.messaging.RMSPDaemon.getActorUAL(ual.getIdentifier()) != null) {
				// Already in this stage...
				remoteReference = wwc.messaging.RMSPDaemon.getActorUAL(ual.getIdentifier());
			}
		}
		// 	start();
	}

	// couldn't "enableResolveObject(true)" in ObjectInputStream
	// so i'm using remoteReference instead as shown above
	protected Object resolveObject(Object obj) throws IOException
	{
		String localHost = InetAddress.getLocalHost().getHostName();
		UAL newual = ((UniversalActor)obj).getUAL();
		if ((newual.getHost().equals(localHost)) && (newual.getPort() == wwc.messaging.RMSPDaemon.getListeningPortNumber())) {
			// reading an already *local* actor.
			return wwc.messaging.RMSPDaemon.getActorUAL(newual.getIdentifier());
		} else
			return obj;
	}
}
